#include "glwidget.h"
#include "plot.h"
#include <iostream>
#include <QTextStream>
#include <QFile>
#include <QIODevice>
#include <math.h>
#include <QMouseEvent>

#define EPSILON .0000001
#define SPH 0
#define TOR 1
#define EXP 2

static void normalize(Xyz *normal);
static int polygonise(Gridcell grid, std::vector <Xyz> &vertex, std::vector <Triangle> &Triangles, int cou[]);
static Xyz VertexInterp(double isolevel, Xyz p1, Xyz p2, double valp1, double valp2, char type);
void init_lang();
void scene_clone_gl(Object *ol1, Object *ol2);


GLWidget::GLWidget(QWidget *parent) : QOpenGLWidget(parent), scene(NULL), auxScene(NULL), expression_str("")
{    
    startTimer( 70 );
    this->rotating = false;
    this->quatStart = this->quatCurrent = QQuaternion();
    this->m = m4_ident();
    this->symbol_table.add_constants();
    this->symbol_table.create_variable("x");
    this->symbol_table.create_variable("y");
    this->symbol_table.create_variable("z");
    this->expression.register_symbol_table(this->symbol_table);
    init_lang();
}

GLWidget::~GLWidget()
{
    delete(scene);
}


void GLWidget::initializeGL()
{
    //initialization of OpenGL

    glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
    glEnable( GL_TEXTURE_2D );
    glEnable( GL_CULL_FACE );


        //initialization of GLEW from s3d
    init_gl(NULL);

    /*dvwindow(740, 480);
    glClearColor(0.0f, 0.0f, 0.0f, 0.f);
    glClear(GL_COLOR_BUFFER_BIT);*/
}

void GLWidget::paintGL()
{
    if (this->auxScene == NULL) {
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    } else {
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        display_gl(this->auxScene);
        scene_free(this->auxScene);
        this->auxScene = NULL;
    }
}

void GLWidget::resizeGL(int width, int height)
{
    dvwindow(width, height);
    glViewport(0, 0, width, height);
    centerOfScreen.setX(width/2.0f);
    centerOfScreen.setY(height/2.0f);
    radiusOfSphere = qMin(centerOfScreen.x(), centerOfScreen.y());
    centerOfScreen.setZ(-radiusOfSphere);
}

QSize GLWidget::sizeHint() const
{
    return QSize(740, 480);
}

void GLWidget::mousePressEvent(QMouseEvent *event)
{
  if (event->button() == Qt::LeftButton) {
      rotating  = true;
      vecStart = projectToSphere(event->x(), event->y());
      quatStart = quatCurrent;

  } else if (event->button() == Qt::RightButton) {
      translating = true;
      vecTransCenter = QVector3D(event->x(), event->y(), 0);

  } else if (event->button() == Qt::MidButton) {
      rotating = true;
      xpos = event->x();
      ypos = event->y();
  }
}

void GLWidget::mouseReleaseEvent(QMouseEvent *event)
{
  if (event->button() == Qt::LeftButton) {
      rotating = false;
    } else if (event->button() == Qt::RightButton) {
      translating = false;
    }
}

void GLWidget::mouseMoveEvent(QMouseEvent *event)
{
    if (rotating) {
      vecCurrent = projectToSphere(event->x(), event->y());

      QVector3D c = QVector3D::crossProduct(vecStart, vecCurrent);
      double    d = QVector3D::dotProduct(vecStart, vecCurrent);

      if (c.length() > EPSILON) {
        quatCurrent.setVector(c);
        quatCurrent.setScalar(d);
        quatCurrent *= quatStart;
      }
      m = quatToMatrix(quatCurrent);

    } else if (translating) {
      QVector3D vecTrans;
      vecTrans.setX((vecTransCenter.x() - event->x()) / this->width() );
      vecTrans.setY((vecTransCenter.y() - event->y()) / this->height() );
      vecTrans.setZ( m.r3.w );

      m = m4_m4prod(m4_translate(vecTrans.x(), vecTrans.y(), vecTrans.z()) , quatToMatrix(quatCurrent));
    }
}

void GLWidget::timerEvent(QTimerEvent *event)
{
    if (this->scene != NULL) {
        if (this->auxScene != NULL){
            close_gl(this->auxScene);
        }
        this->auxScene = scene_eval();
        this->auxScene->view->V = m4_m4prod(this->auxScene->view->V, this->m);
        //scene_clone_gl(this->auxScene->objs, this->scene->objs);
        load_objects_gl(this->auxScene->objs);
        setup_programs_gl(this->auxScene, this->auxScene->objs);
        time_incr(1);
    }
    this->update();
}

QVector3D GLWidget::projectToSphere(double x, double y)
{
    QVector3D vecClicked(x,y,0);
    QVector3D centerXY = centerOfScreen;
    centerXY.setZ(0);

    vecClicked.setX((x - centerOfScreen.x()) / radiusOfSphere);
    vecClicked.setY((y - centerOfScreen.y()) / radiusOfSphere);

    double r =  vecClicked.x()*vecClicked.x() + vecClicked.y()*vecClicked.y();
    if (r < 1.0)
    {
      vecClicked.setZ(sqrt(1. - r));
    } else
    {
      double s = 1/sqrt(r);
      vecClicked.setX(vecClicked.x()*s);
      vecClicked.setY(vecClicked.y()*s);
      vecClicked.setZ(0.);
    }

    return vecClicked;
}

Matrix4 GLWidget::quatToMatrix(QQuaternion &quat)
{
  QMatrix4x4 m;
  m.rotate(quat);

  Matrix4 matrix;
  matrix.r1 = v4_make(m.row(0).x(), m.row(0).y(), m.row(0).z(), m.row(0).w());
  matrix.r2 = v4_make(m.row(1).x(), m.row(1).y(), m.row(1).z(), m.row(1).w());
  matrix.r3 = v4_make(m.row(2).x(), m.row(2).y(), m.row(2).z(), m.row(2).w());
  matrix.r4 = v4_make(m.row(3).x(), m.row(3).y(), m.row(3).z(), m.row(3).w());

  return matrix;
}

void GLWidget::dragEnterEvent(QDragEnterEvent *event)
{
    event->acceptProposedAction();
}

void GLWidget::dragMoveEvent(QDragMoveEvent *event)
{
    event->acceptProposedAction();
}


void GLWidget::dragLeaveEvent(QDragLeaveEvent *event)
{
    event->accept();
}

void GLWidget::keyPressEvent(QKeyEvent *event)
{
    switch (event->key()) {
    case Qt::Key_Q:
    case Qt::Key_Escape:
           exit(0);
           break;
    case Qt::Key_R:
        break;
    }
    update();
}

void GLWidget::dropEvent(QDropEvent *event)
{

}

void GLWidget::renderCloud(){
    int n;
    QString line;
    QFile input("/home/pabloant/Desktop/ComputerGraphics/s3d-v53/platforms/qt/marching/marching/rhino.off");
    QFile output("/home/pabloant/Desktop/ComputerGraphics/s3d-v53/platforms/qt/marching/marching/output.scn");
    if (input.open(QFile::ReadOnly) && output.open(QFile::WriteOnly)) {
        QTextStream out(&output);
        QTextStream in(&input);

        out << "scene{\n" << " camera = view { from = {0, 0.5, -1.5}, at = {0, 0, 0}, up = {0, 1, 0}, fov = 60, imgw = 640, imgh = 480},\n";
        out << " light = dist_light { direction = {0, 1, 1} },\n";
        out << " light = dist_light { direction = {0, -1, -1} },\n";

        in.readLine();
        line = in.readLine();
        QStringList pieces = line.split(" ");
        if ( pieces.value(0).toInt() > 0){
            n = pieces.value(0).toInt();

            for (int i = 0; i<n; i++){
                line = in.readLine();
                pieces = line.split(" ");

                out << "object = primobj {shape = sphere {radius = .0000005, center = {" << pieces.value(0) << ", " << pieces.value(1) <<
                       ", " << pieces.value(2) << "} }},\n";


            }
            out << "};";
        }
        input.close();
        output.close();
    }
}

void GLWidget::exampleMarching(int type, int gridSize, int *nv, int *nt, double xMin, double yMin, double zMin, double xMax, double yMax, double zMax, double fromX
                               , double fromY, double fromZ, double fov){
    double deltaX = (xMax-xMin)/((double)(gridSize-1));

    Gridcell grid;
    std::vector <Xyz> vertex(15);
    std::vector <Triangle> Triangles(10);
    int count[2] = {0,0};

    for (int i = 0; i<gridSize; i++){
        grid.p[0].x = xMin + i*deltaX; grid.p[1].x = xMin + (i+1)*deltaX; grid.p[2].x = xMin +(i+1)*deltaX; grid.p[3].x = xMin +i*deltaX;
        grid.p[4].x = xMin +i*deltaX; grid.p[5].x = xMin +(i+1)*deltaX; grid.p[6].x = xMin +(i+1)*deltaX; grid.p[7].x = xMin +i*deltaX;
        for (int j = 0; j<gridSize; j++){
            grid.p[0].y = yMin +j*deltaX; grid.p[1].y = yMin +j*deltaX; grid.p[2].y = yMin +(j+1)*deltaX; grid.p[3].y =yMin + (j+1)*deltaX;
            grid.p[4].y =yMin + j*deltaX; grid.p[5].y = yMin +j*deltaX; grid.p[6].y =yMin + (j+1)*deltaX; grid.p[7].y = yMin +(j+1)*deltaX;
            for (int k = 0; k<gridSize; k++){

                grid.p[0].z =zMin + k*deltaX; grid.p[1].z = zMin +k*deltaX; grid.p[2].z = zMin +k*deltaX; grid.p[3].z = zMin +k*deltaX;
                grid.p[4].z = zMin +(k+1)*deltaX; grid.p[5].z = zMin +(k+1)*deltaX; grid.p[6].z =zMin + (k+1)*deltaX; grid.p[7].z =zMin + (k+1)*deltaX;

                grid.val[0] = functionXyz(grid.p[0].x, grid.p[0].y, grid.p[0].z, type);
                grid.val[1] = functionXyz(grid.p[1].x, grid.p[1].y, grid.p[1].z, type);
                grid.val[2] = functionXyz(grid.p[2].x, grid.p[2].y, grid.p[2].z, type);
                grid.val[3] = functionXyz(grid.p[3].x, grid.p[3].y, grid.p[3].z, type);
                grid.val[4] = functionXyz(grid.p[4].x, grid.p[4].y, grid.p[4].z, type);
                grid.val[5] = functionXyz(grid.p[5].x, grid.p[5].y, grid.p[5].z, type);
                grid.val[6] = functionXyz(grid.p[6].x, grid.p[6].y, grid.p[6].z, type);
                grid.val[7] = functionXyz(grid.p[7].x, grid.p[7].y, grid.p[7].z, type);

                polygonise(grid, vertex, Triangles, count);
            }
        }
    }

    QFile data("/home/pabloant/Desktop/ComputerGraphics/s3d-v53/platforms/qt/marching/marching/output.txt");
    if (data.open(QFile::WriteOnly)) {
        QTextStream out(&data);
        out << "scene{\n" << " camera = view { from = {" << fromX << ", " << fromY << ", " << fromZ << "}, at = {0, 0, 0}, up = {0, 1, 0}, fov = " << fov
            << ", imgw = 640, imgh = 480},\n";
        out << " light = dist_light { direction = {0, 1, 1} },\n";
        out << " light = dist_light { direction = {0, -1, -1} },\n";        
        out << " object = meshobj { material = plastic { d_col = {.79, .69, .9}, shader = \"phong_plastic\"}, shape = mesh {\n";
        out << " points = {";

        for (int i = 0; i<count[0]; i++){
            out << "{ " << vertex[i].x << ", " << vertex[i].y << ", " << vertex[i].z << "},\n";
        }
        out << "},\n faces = {";

        for (int i = 0; i<count[1]; i++){
            out << "{ " << Triangles[i].p[0] << ", " << Triangles[i].p[1] << ", " << Triangles[i].p[2] << "},\n";
        }
        out << "}\n}\n}\n};";
        data.close();
    }
    vertex.resize(0);
    Triangles.resize(0);
    *nv = count[0];
    *nt = count[1];

}

bool GLWidget::parseExpression(){
      parser_t parser;

      if (!parser.compile(expression_str,this->expression))
      {
         fprintf(stderr,"Error: %s\tExpression: %s\n",
                parser.error().c_str(),
                expression_str.c_str());

         for (std::size_t i = 0; i < parser.error_count(); ++i)
         {
            error_t error = parser.get_error(i);
            fprintf(stderr,"Error: %02d Position: %02d Type: [%s] Msg: %s Expr: %s\n",
                   static_cast<int>(i),
                   static_cast<int>(error.token.position),
                   exprtk::parser_error::to_str(error.mode).c_str(),
                   error.diagnostic.c_str(),
                   expression_str.c_str());
         }
         return false;
      }
      return true;
}

double GLWidget::evaluateExpression(double x, double y, double z){
    this->symbol_table.get_variable("x")->ref() = x;
    this->symbol_table.get_variable("y")->ref() = y;
    this->symbol_table.get_variable("z")->ref() = z;

    return this->expression.value();

    /*double R = 0.3;
    return x*x+y*y+z*z-R*R; // sphere */
}

double GLWidget::functionXyz(double x, double y, double z, int type){
    double R = 0.3;
    double r = 0.1;

    if (type == TOR)
        return (x*x+y*y+z*z + R*R-r*r)*(x*x+y*y+z*z + R*R-r*r) - 4*R*R*(x*x+y*y); // torus
    else if (type == SPH)
        return x*x+y*y+z*z-R*R; // sphere
    else if (type == EXP){
        return evaluateExpression(x, y, z);
    }
    else
        return 0;
}

// -------------------------------------------------------------------------------

static int polygonise(Gridcell grid, std::vector <Xyz> &vertex, std::vector <Triangle> &Triangles, int cou[])
{
   int i,nTriang, nVertex;
   int cubeindex;
   int vertlist[12];

int edgeTable[256]={
0x0  , 0x109, 0x203, 0x30a, 0x406, 0x50f, 0x605, 0x70c,
0x80c, 0x905, 0xa0f, 0xb06, 0xc0a, 0xd03, 0xe09, 0xf00,
0x190, 0x99 , 0x393, 0x29a, 0x596, 0x49f, 0x795, 0x69c,
0x99c, 0x895, 0xb9f, 0xa96, 0xd9a, 0xc93, 0xf99, 0xe90,
0x230, 0x339, 0x33 , 0x13a, 0x636, 0x73f, 0x435, 0x53c,
0xa3c, 0xb35, 0x83f, 0x936, 0xe3a, 0xf33, 0xc39, 0xd30,
0x3a0, 0x2a9, 0x1a3, 0xaa , 0x7a6, 0x6af, 0x5a5, 0x4ac,
0xbac, 0xaa5, 0x9af, 0x8a6, 0xfaa, 0xea3, 0xda9, 0xca0,
0x460, 0x569, 0x663, 0x76a, 0x66 , 0x16f, 0x265, 0x36c,
0xc6c, 0xd65, 0xe6f, 0xf66, 0x86a, 0x963, 0xa69, 0xb60,
0x5f0, 0x4f9, 0x7f3, 0x6fa, 0x1f6, 0xff , 0x3f5, 0x2fc,
0xdfc, 0xcf5, 0xfff, 0xef6, 0x9fa, 0x8f3, 0xbf9, 0xaf0,
0x650, 0x759, 0x453, 0x55a, 0x256, 0x35f, 0x55 , 0x15c,
0xe5c, 0xf55, 0xc5f, 0xd56, 0xa5a, 0xb53, 0x859, 0x950,
0x7c0, 0x6c9, 0x5c3, 0x4ca, 0x3c6, 0x2cf, 0x1c5, 0xcc ,
0xfcc, 0xec5, 0xdcf, 0xcc6, 0xbca, 0xac3, 0x9c9, 0x8c0,
0x8c0, 0x9c9, 0xac3, 0xbca, 0xcc6, 0xdcf, 0xec5, 0xfcc,
0xcc , 0x1c5, 0x2cf, 0x3c6, 0x4ca, 0x5c3, 0x6c9, 0x7c0,
0x950, 0x859, 0xb53, 0xa5a, 0xd56, 0xc5f, 0xf55, 0xe5c,
0x15c, 0x55 , 0x35f, 0x256, 0x55a, 0x453, 0x759, 0x650,
0xaf0, 0xbf9, 0x8f3, 0x9fa, 0xef6, 0xfff, 0xcf5, 0xdfc,
0x2fc, 0x3f5, 0xff , 0x1f6, 0x6fa, 0x7f3, 0x4f9, 0x5f0,
0xb60, 0xa69, 0x963, 0x86a, 0xf66, 0xe6f, 0xd65, 0xc6c,
0x36c, 0x265, 0x16f, 0x66 , 0x76a, 0x663, 0x569, 0x460,
0xca0, 0xda9, 0xea3, 0xfaa, 0x8a6, 0x9af, 0xaa5, 0xbac,
0x4ac, 0x5a5, 0x6af, 0x7a6, 0xaa , 0x1a3, 0x2a9, 0x3a0,
0xd30, 0xc39, 0xf33, 0xe3a, 0x936, 0x83f, 0xb35, 0xa3c,
0x53c, 0x435, 0x73f, 0x636, 0x13a, 0x33 , 0x339, 0x230,
0xe90, 0xf99, 0xc93, 0xd9a, 0xa96, 0xb9f, 0x895, 0x99c,
0x69c, 0x795, 0x49f, 0x596, 0x29a, 0x393, 0x99 , 0x190,
0xf00, 0xe09, 0xd03, 0xc0a, 0xb06, 0xa0f, 0x905, 0x80c,
0x70c, 0x605, 0x50f, 0x406, 0x30a, 0x203, 0x109, 0x0   };
int triTable[256][16] =
{{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 1, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 8, 3, 9, 8, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 3, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{9, 2, 10, 0, 2, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{2, 8, 3, 2, 10, 8, 10, 9, 8, -1, -1, -1, -1, -1, -1, -1},
{3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 11, 2, 8, 11, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 9, 0, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 11, 2, 1, 9, 11, 9, 8, 11, -1, -1, -1, -1, -1, -1, -1},
{3, 10, 1, 11, 10, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 10, 1, 0, 8, 10, 8, 11, 10, -1, -1, -1, -1, -1, -1, -1},
{3, 9, 0, 3, 11, 9, 11, 10, 9, -1, -1, -1, -1, -1, -1, -1},
{9, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 3, 0, 7, 3, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 1, 9, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 1, 9, 4, 7, 1, 7, 3, 1, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 10, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{3, 4, 7, 3, 0, 4, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1},
{9, 2, 10, 9, 0, 2, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
{2, 10, 9, 2, 9, 7, 2, 7, 3, 7, 9, 4, -1, -1, -1, -1},
{8, 4, 7, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{11, 4, 7, 11, 2, 4, 2, 0, 4, -1, -1, -1, -1, -1, -1, -1},
{9, 0, 1, 8, 4, 7, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
{4, 7, 11, 9, 4, 11, 9, 11, 2, 9, 2, 1, -1, -1, -1, -1},
{3, 10, 1, 3, 11, 10, 7, 8, 4, -1, -1, -1, -1, -1, -1, -1},
{1, 11, 10, 1, 4, 11, 1, 0, 4, 7, 11, 4, -1, -1, -1, -1},
{4, 7, 8, 9, 0, 11, 9, 11, 10, 11, 0, 3, -1, -1, -1, -1},
{4, 7, 11, 4, 11, 9, 9, 11, 10, -1, -1, -1, -1, -1, -1, -1},
{9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{9, 5, 4, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 5, 4, 1, 5, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{8, 5, 4, 8, 3, 5, 3, 1, 5, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 10, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{3, 0, 8, 1, 2, 10, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
{5, 2, 10, 5, 4, 2, 4, 0, 2, -1, -1, -1, -1, -1, -1, -1},
{2, 10, 5, 3, 2, 5, 3, 5, 4, 3, 4, 8, -1, -1, -1, -1},
{9, 5, 4, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 11, 2, 0, 8, 11, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
{0, 5, 4, 0, 1, 5, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
{2, 1, 5, 2, 5, 8, 2, 8, 11, 4, 8, 5, -1, -1, -1, -1},
{10, 3, 11, 10, 1, 3, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1},
{4, 9, 5, 0, 8, 1, 8, 10, 1, 8, 11, 10, -1, -1, -1, -1},
{5, 4, 0, 5, 0, 11, 5, 11, 10, 11, 0, 3, -1, -1, -1, -1},
{5, 4, 8, 5, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1},
{9, 7, 8, 5, 7, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{9, 3, 0, 9, 5, 3, 5, 7, 3, -1, -1, -1, -1, -1, -1, -1},
{0, 7, 8, 0, 1, 7, 1, 5, 7, -1, -1, -1, -1, -1, -1, -1},
{1, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{9, 7, 8, 9, 5, 7, 10, 1, 2, -1, -1, -1, -1, -1, -1, -1},
{10, 1, 2, 9, 5, 0, 5, 3, 0, 5, 7, 3, -1, -1, -1, -1},
{8, 0, 2, 8, 2, 5, 8, 5, 7, 10, 5, 2, -1, -1, -1, -1},
{2, 10, 5, 2, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1},
{7, 9, 5, 7, 8, 9, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1},
{9, 5, 7, 9, 7, 2, 9, 2, 0, 2, 7, 11, -1, -1, -1, -1},
{2, 3, 11, 0, 1, 8, 1, 7, 8, 1, 5, 7, -1, -1, -1, -1},
{11, 2, 1, 11, 1, 7, 7, 1, 5, -1, -1, -1, -1, -1, -1, -1},
{9, 5, 8, 8, 5, 7, 10, 1, 3, 10, 3, 11, -1, -1, -1, -1},
{5, 7, 0, 5, 0, 9, 7, 11, 0, 1, 0, 10, 11, 10, 0, -1},
{11, 10, 0, 11, 0, 3, 10, 5, 0, 8, 0, 7, 5, 7, 0, -1},
{11, 10, 5, 7, 11, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 3, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{9, 0, 1, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 8, 3, 1, 9, 8, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
{1, 6, 5, 2, 6, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 6, 5, 1, 2, 6, 3, 0, 8, -1, -1, -1, -1, -1, -1, -1},
{9, 6, 5, 9, 0, 6, 0, 2, 6, -1, -1, -1, -1, -1, -1, -1},
{5, 9, 8, 5, 8, 2, 5, 2, 6, 3, 2, 8, -1, -1, -1, -1},
{2, 3, 11, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{11, 0, 8, 11, 2, 0, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
{0, 1, 9, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
{5, 10, 6, 1, 9, 2, 9, 11, 2, 9, 8, 11, -1, -1, -1, -1},
{6, 3, 11, 6, 5, 3, 5, 1, 3, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 11, 0, 11, 5, 0, 5, 1, 5, 11, 6, -1, -1, -1, -1},
{3, 11, 6, 0, 3, 6, 0, 6, 5, 0, 5, 9, -1, -1, -1, -1},
{6, 5, 9, 6, 9, 11, 11, 9, 8, -1, -1, -1, -1, -1, -1, -1},
{5, 10, 6, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 3, 0, 4, 7, 3, 6, 5, 10, -1, -1, -1, -1, -1, -1, -1},
{1, 9, 0, 5, 10, 6, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
{10, 6, 5, 1, 9, 7, 1, 7, 3, 7, 9, 4, -1, -1, -1, -1},
{6, 1, 2, 6, 5, 1, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 5, 5, 2, 6, 3, 0, 4, 3, 4, 7, -1, -1, -1, -1},
{8, 4, 7, 9, 0, 5, 0, 6, 5, 0, 2, 6, -1, -1, -1, -1},
{7, 3, 9, 7, 9, 4, 3, 2, 9, 5, 9, 6, 2, 6, 9, -1},
{3, 11, 2, 7, 8, 4, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
{5, 10, 6, 4, 7, 2, 4, 2, 0, 2, 7, 11, -1, -1, -1, -1},
{0, 1, 9, 4, 7, 8, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1},
{9, 2, 1, 9, 11, 2, 9, 4, 11, 7, 11, 4, 5, 10, 6, -1},
{8, 4, 7, 3, 11, 5, 3, 5, 1, 5, 11, 6, -1, -1, -1, -1},
{5, 1, 11, 5, 11, 6, 1, 0, 11, 7, 11, 4, 0, 4, 11, -1},
{0, 5, 9, 0, 6, 5, 0, 3, 6, 11, 6, 3, 8, 4, 7, -1},
{6, 5, 9, 6, 9, 11, 4, 7, 9, 7, 11, 9, -1, -1, -1, -1},
{10, 4, 9, 6, 4, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 10, 6, 4, 9, 10, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1},
{10, 0, 1, 10, 6, 0, 6, 4, 0, -1, -1, -1, -1, -1, -1, -1},
{8, 3, 1, 8, 1, 6, 8, 6, 4, 6, 1, 10, -1, -1, -1, -1},
{1, 4, 9, 1, 2, 4, 2, 6, 4, -1, -1, -1, -1, -1, -1, -1},
{3, 0, 8, 1, 2, 9, 2, 4, 9, 2, 6, 4, -1, -1, -1, -1},
{0, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{8, 3, 2, 8, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1},
{10, 4, 9, 10, 6, 4, 11, 2, 3, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 2, 2, 8, 11, 4, 9, 10, 4, 10, 6, -1, -1, -1, -1},
{3, 11, 2, 0, 1, 6, 0, 6, 4, 6, 1, 10, -1, -1, -1, -1},
{6, 4, 1, 6, 1, 10, 4, 8, 1, 2, 1, 11, 8, 11, 1, -1},
{9, 6, 4, 9, 3, 6, 9, 1, 3, 11, 6, 3, -1, -1, -1, -1},
{8, 11, 1, 8, 1, 0, 11, 6, 1, 9, 1, 4, 6, 4, 1, -1},
{3, 11, 6, 3, 6, 0, 0, 6, 4, -1, -1, -1, -1, -1, -1, -1},
{6, 4, 8, 11, 6, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{7, 10, 6, 7, 8, 10, 8, 9, 10, -1, -1, -1, -1, -1, -1, -1},
{0, 7, 3, 0, 10, 7, 0, 9, 10, 6, 7, 10, -1, -1, -1, -1},
{10, 6, 7, 1, 10, 7, 1, 7, 8, 1, 8, 0, -1, -1, -1, -1},
{10, 6, 7, 10, 7, 1, 1, 7, 3, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 6, 1, 6, 8, 1, 8, 9, 8, 6, 7, -1, -1, -1, -1},
{2, 6, 9, 2, 9, 1, 6, 7, 9, 0, 9, 3, 7, 3, 9, -1},
{7, 8, 0, 7, 0, 6, 6, 0, 2, -1, -1, -1, -1, -1, -1, -1},
{7, 3, 2, 6, 7, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{2, 3, 11, 10, 6, 8, 10, 8, 9, 8, 6, 7, -1, -1, -1, -1},
{2, 0, 7, 2, 7, 11, 0, 9, 7, 6, 7, 10, 9, 10, 7, -1},
{1, 8, 0, 1, 7, 8, 1, 10, 7, 6, 7, 10, 2, 3, 11, -1},
{11, 2, 1, 11, 1, 7, 10, 6, 1, 6, 7, 1, -1, -1, -1, -1},
{8, 9, 6, 8, 6, 7, 9, 1, 6, 11, 6, 3, 1, 3, 6, -1},
{0, 9, 1, 11, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{7, 8, 0, 7, 0, 6, 3, 11, 0, 11, 6, 0, -1, -1, -1, -1},
{7, 11, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{3, 0, 8, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 1, 9, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{8, 1, 9, 8, 3, 1, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
{10, 1, 2, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 10, 3, 0, 8, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
{2, 9, 0, 2, 10, 9, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
{6, 11, 7, 2, 10, 3, 10, 8, 3, 10, 9, 8, -1, -1, -1, -1},
{7, 2, 3, 6, 2, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{7, 0, 8, 7, 6, 0, 6, 2, 0, -1, -1, -1, -1, -1, -1, -1},
{2, 7, 6, 2, 3, 7, 0, 1, 9, -1, -1, -1, -1, -1, -1, -1},
{1, 6, 2, 1, 8, 6, 1, 9, 8, 8, 7, 6, -1, -1, -1, -1},
{10, 7, 6, 10, 1, 7, 1, 3, 7, -1, -1, -1, -1, -1, -1, -1},
{10, 7, 6, 1, 7, 10, 1, 8, 7, 1, 0, 8, -1, -1, -1, -1},
{0, 3, 7, 0, 7, 10, 0, 10, 9, 6, 10, 7, -1, -1, -1, -1},
{7, 6, 10, 7, 10, 8, 8, 10, 9, -1, -1, -1, -1, -1, -1, -1},
{6, 8, 4, 11, 8, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{3, 6, 11, 3, 0, 6, 0, 4, 6, -1, -1, -1, -1, -1, -1, -1},
{8, 6, 11, 8, 4, 6, 9, 0, 1, -1, -1, -1, -1, -1, -1, -1},
{9, 4, 6, 9, 6, 3, 9, 3, 1, 11, 3, 6, -1, -1, -1, -1},
{6, 8, 4, 6, 11, 8, 2, 10, 1, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 10, 3, 0, 11, 0, 6, 11, 0, 4, 6, -1, -1, -1, -1},
{4, 11, 8, 4, 6, 11, 0, 2, 9, 2, 10, 9, -1, -1, -1, -1},
{10, 9, 3, 10, 3, 2, 9, 4, 3, 11, 3, 6, 4, 6, 3, -1},
{8, 2, 3, 8, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1},
{0, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 9, 0, 2, 3, 4, 2, 4, 6, 4, 3, 8, -1, -1, -1, -1},
{1, 9, 4, 1, 4, 2, 2, 4, 6, -1, -1, -1, -1, -1, -1, -1},
{8, 1, 3, 8, 6, 1, 8, 4, 6, 6, 10, 1, -1, -1, -1, -1},
{10, 1, 0, 10, 0, 6, 6, 0, 4, -1, -1, -1, -1, -1, -1, -1},
{4, 6, 3, 4, 3, 8, 6, 10, 3, 0, 3, 9, 10, 9, 3, -1},
{10, 9, 4, 6, 10, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 9, 5, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 3, 4, 9, 5, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
{5, 0, 1, 5, 4, 0, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
{11, 7, 6, 8, 3, 4, 3, 5, 4, 3, 1, 5, -1, -1, -1, -1},
{9, 5, 4, 10, 1, 2, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
{6, 11, 7, 1, 2, 10, 0, 8, 3, 4, 9, 5, -1, -1, -1, -1},
{7, 6, 11, 5, 4, 10, 4, 2, 10, 4, 0, 2, -1, -1, -1, -1},
{3, 4, 8, 3, 5, 4, 3, 2, 5, 10, 5, 2, 11, 7, 6, -1},
{7, 2, 3, 7, 6, 2, 5, 4, 9, -1, -1, -1, -1, -1, -1, -1},
{9, 5, 4, 0, 8, 6, 0, 6, 2, 6, 8, 7, -1, -1, -1, -1},
{3, 6, 2, 3, 7, 6, 1, 5, 0, 5, 4, 0, -1, -1, -1, -1},
{6, 2, 8, 6, 8, 7, 2, 1, 8, 4, 8, 5, 1, 5, 8, -1},
{9, 5, 4, 10, 1, 6, 1, 7, 6, 1, 3, 7, -1, -1, -1, -1},
{1, 6, 10, 1, 7, 6, 1, 0, 7, 8, 7, 0, 9, 5, 4, -1},
{4, 0, 10, 4, 10, 5, 0, 3, 10, 6, 10, 7, 3, 7, 10, -1},
{7, 6, 10, 7, 10, 8, 5, 4, 10, 4, 8, 10, -1, -1, -1, -1},
{6, 9, 5, 6, 11, 9, 11, 8, 9, -1, -1, -1, -1, -1, -1, -1},
{3, 6, 11, 0, 6, 3, 0, 5, 6, 0, 9, 5, -1, -1, -1, -1},
{0, 11, 8, 0, 5, 11, 0, 1, 5, 5, 6, 11, -1, -1, -1, -1},
{6, 11, 3, 6, 3, 5, 5, 3, 1, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 10, 9, 5, 11, 9, 11, 8, 11, 5, 6, -1, -1, -1, -1},
{0, 11, 3, 0, 6, 11, 0, 9, 6, 5, 6, 9, 1, 2, 10, -1},
{11, 8, 5, 11, 5, 6, 8, 0, 5, 10, 5, 2, 0, 2, 5, -1},
{6, 11, 3, 6, 3, 5, 2, 10, 3, 10, 5, 3, -1, -1, -1, -1},
{5, 8, 9, 5, 2, 8, 5, 6, 2, 3, 8, 2, -1, -1, -1, -1},
{9, 5, 6, 9, 6, 0, 0, 6, 2, -1, -1, -1, -1, -1, -1, -1},
{1, 5, 8, 1, 8, 0, 5, 6, 8, 3, 8, 2, 6, 2, 8, -1},
{1, 5, 6, 2, 1, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 3, 6, 1, 6, 10, 3, 8, 6, 5, 6, 9, 8, 9, 6, -1},
{10, 1, 0, 10, 0, 6, 9, 5, 0, 5, 6, 0, -1, -1, -1, -1},
{0, 3, 8, 5, 6, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{10, 5, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{11, 5, 10, 7, 5, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{11, 5, 10, 11, 7, 5, 8, 3, 0, -1, -1, -1, -1, -1, -1, -1},
{5, 11, 7, 5, 10, 11, 1, 9, 0, -1, -1, -1, -1, -1, -1, -1},
{10, 7, 5, 10, 11, 7, 9, 8, 1, 8, 3, 1, -1, -1, -1, -1},
{11, 1, 2, 11, 7, 1, 7, 5, 1, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 3, 1, 2, 7, 1, 7, 5, 7, 2, 11, -1, -1, -1, -1},
{9, 7, 5, 9, 2, 7, 9, 0, 2, 2, 11, 7, -1, -1, -1, -1},
{7, 5, 2, 7, 2, 11, 5, 9, 2, 3, 2, 8, 9, 8, 2, -1},
{2, 5, 10, 2, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1},
{8, 2, 0, 8, 5, 2, 8, 7, 5, 10, 2, 5, -1, -1, -1, -1},
{9, 0, 1, 5, 10, 3, 5, 3, 7, 3, 10, 2, -1, -1, -1, -1},
{9, 8, 2, 9, 2, 1, 8, 7, 2, 10, 2, 5, 7, 5, 2, -1},
{1, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 7, 0, 7, 1, 1, 7, 5, -1, -1, -1, -1, -1, -1, -1},
{9, 0, 3, 9, 3, 5, 5, 3, 7, -1, -1, -1, -1, -1, -1, -1},
{9, 8, 7, 5, 9, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{5, 8, 4, 5, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1},
{5, 0, 4, 5, 11, 0, 5, 10, 11, 11, 3, 0, -1, -1, -1, -1},
{0, 1, 9, 8, 4, 10, 8, 10, 11, 10, 4, 5, -1, -1, -1, -1},
{10, 11, 4, 10, 4, 5, 11, 3, 4, 9, 4, 1, 3, 1, 4, -1},
{2, 5, 1, 2, 8, 5, 2, 11, 8, 4, 5, 8, -1, -1, -1, -1},
{0, 4, 11, 0, 11, 3, 4, 5, 11, 2, 11, 1, 5, 1, 11, -1},
{0, 2, 5, 0, 5, 9, 2, 11, 5, 4, 5, 8, 11, 8, 5, -1},
{9, 4, 5, 2, 11, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{2, 5, 10, 3, 5, 2, 3, 4, 5, 3, 8, 4, -1, -1, -1, -1},
{5, 10, 2, 5, 2, 4, 4, 2, 0, -1, -1, -1, -1, -1, -1, -1},
{3, 10, 2, 3, 5, 10, 3, 8, 5, 4, 5, 8, 0, 1, 9, -1},
{5, 10, 2, 5, 2, 4, 1, 9, 2, 9, 4, 2, -1, -1, -1, -1},
{8, 4, 5, 8, 5, 3, 3, 5, 1, -1, -1, -1, -1, -1, -1, -1},
{0, 4, 5, 1, 0, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{8, 4, 5, 8, 5, 3, 9, 0, 5, 0, 3, 5, -1, -1, -1, -1},
{9, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 11, 7, 4, 9, 11, 9, 10, 11, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 3, 4, 9, 7, 9, 11, 7, 9, 10, 11, -1, -1, -1, -1},
{1, 10, 11, 1, 11, 4, 1, 4, 0, 7, 4, 11, -1, -1, -1, -1},
{3, 1, 4, 3, 4, 8, 1, 10, 4, 7, 4, 11, 10, 11, 4, -1},
{4, 11, 7, 9, 11, 4, 9, 2, 11, 9, 1, 2, -1, -1, -1, -1},
{9, 7, 4, 9, 11, 7, 9, 1, 11, 2, 11, 1, 0, 8, 3, -1},
{11, 7, 4, 11, 4, 2, 2, 4, 0, -1, -1, -1, -1, -1, -1, -1},
{11, 7, 4, 11, 4, 2, 8, 3, 4, 3, 2, 4, -1, -1, -1, -1},
{2, 9, 10, 2, 7, 9, 2, 3, 7, 7, 4, 9, -1, -1, -1, -1},
{9, 10, 7, 9, 7, 4, 10, 2, 7, 8, 7, 0, 2, 0, 7, -1},
{3, 7, 10, 3, 10, 2, 7, 4, 10, 1, 10, 0, 4, 0, 10, -1},
{1, 10, 2, 8, 7, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 9, 1, 4, 1, 7, 7, 1, 3, -1, -1, -1, -1, -1, -1, -1},
{4, 9, 1, 4, 1, 7, 0, 8, 1, 8, 7, 1, -1, -1, -1, -1},
{4, 0, 3, 7, 4, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 8, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{9, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{3, 0, 9, 3, 9, 11, 11, 9, 10, -1, -1, -1, -1, -1, -1, -1},
{0, 1, 10, 0, 10, 8, 8, 10, 11, -1, -1, -1, -1, -1, -1, -1},
{3, 1, 10, 11, 3, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 11, 1, 11, 9, 9, 11, 8, -1, -1, -1, -1, -1, -1, -1},
{3, 0, 9, 3, 9, 11, 1, 2, 9, 2, 11, 9, -1, -1, -1, -1},
{0, 2, 11, 8, 0, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{3, 2, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{2, 3, 8, 2, 8, 10, 10, 8, 9, -1, -1, -1, -1, -1, -1, -1},
{9, 10, 2, 0, 9, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{2, 3, 8, 2, 8, 10, 0, 1, 8, 1, 10, 8, -1, -1, -1, -1},
{1, 10, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 3, 8, 9, 1, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 9, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 3, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}};

   /*
      Determine the index into the edge table which
      tells us which vertices are inside of the surface
   */
   cubeindex = 0;
   if (grid.val[0] <= 0) cubeindex |= 1;
   if (grid.val[1] <= 0) cubeindex |= 2;
   if (grid.val[2] <= 0) cubeindex |= 4;
   if (grid.val[3] <= 0) cubeindex |= 8;
   if (grid.val[4] <= 0) cubeindex |= 16;
   if (grid.val[5] <= 0) cubeindex |= 32;
   if (grid.val[6] <= 0) cubeindex |= 64;
   if (grid.val[7] <= 0) cubeindex |= 128;

   /* Cube is entirely in/out of the surface */
   if (edgeTable[cubeindex] == 0)
       return 0;

   double isolevel = 0;
   /* Find the vertices where the surface intersects the cube */
   nVertex = cou[0];
   if ((edgeTable[cubeindex] & 1) == 1){
        if (vertex.size() < nVertex+1 )
            vertex.resize( 2*vertex.size() );
        vertlist[0] = nVertex;
        vertex[nVertex] = VertexInterp(isolevel,grid.p[0],grid.p[1],grid.val[0],grid.val[1],'X');
        nVertex++;
   }
    if ((edgeTable[cubeindex] & 2) == 2){
        if (vertex.size() < nVertex+1 )
            vertex.resize( 2*vertex.size() );
        vertlist[1] = nVertex;
        vertex[nVertex] = VertexInterp(isolevel,grid.p[1],grid.p[2],grid.val[1],grid.val[2],'Y');
        nVertex++;
    }

   if ((edgeTable[cubeindex] & 4) == 4){
       if (vertex.size() < nVertex+1 )
           vertex.resize( 2*vertex.size() );
       vertlist[2] = nVertex;
       vertex[nVertex] = VertexInterp(isolevel,grid.p[2],grid.p[3],grid.val[2],grid.val[3],'X');
        nVertex++;
   }

   if ((edgeTable[cubeindex] & 8) == 8){
      if (vertex.size() < nVertex+1 )
          vertex.resize( 2*vertex.size() );
      vertlist[3] = nVertex;
      vertex[nVertex] = VertexInterp(isolevel,grid.p[3],grid.p[0],grid.val[3],grid.val[0],'Y');
      nVertex++;
   }
   if ((edgeTable[cubeindex] & 16) == 16){
      if (vertex.size() < nVertex+1 )
          vertex.resize( 2*vertex.size() );
      vertlist[4] = nVertex;
      vertex[nVertex] = VertexInterp(isolevel,grid.p[4],grid.p[5],grid.val[4],grid.val[5],'X');
      nVertex++;
   }
   if ((edgeTable[cubeindex] & 32) == 32){
      if (vertex.size() < nVertex+1 )
          vertex.resize( 2*vertex.size() );
      vertlist[5] = nVertex;
      vertex[nVertex] = VertexInterp(isolevel,grid.p[5],grid.p[6],grid.val[5],grid.val[6],'Y');
      nVertex++;
   }
   if ((edgeTable[cubeindex] & 64) == 64){
      if (vertex.size() < nVertex+1 )
          vertex.resize( 2*vertex.size() );
      vertlist[6] = nVertex;
      vertex[nVertex] = VertexInterp(isolevel,grid.p[6],grid.p[7],grid.val[6],grid.val[7],'X');
      nVertex++;
   }
   if ((edgeTable[cubeindex] & 128) == 128){
      if (vertex.size() < nVertex+1 )
          vertex.resize( 2*vertex.size() );
      vertlist[7] = nVertex;
      vertex[nVertex] = VertexInterp(isolevel,grid.p[7],grid.p[4],grid.val[7],grid.val[4],'Y');
      nVertex++;
   }
   if ((edgeTable[cubeindex] & 256) == 256){
      if (vertex.size() < nVertex+1 )
          vertex.resize( 2*vertex.size() );
      vertlist[8] = nVertex;
      vertex[nVertex] = VertexInterp(isolevel,grid.p[0],grid.p[4],grid.val[0],grid.val[4],'Z');
      nVertex++;
   }
   if ((edgeTable[cubeindex] & 512) == 512){
      if (vertex.size() < nVertex+1 )
          vertex.resize( 2*vertex.size() );
      vertlist[9] = nVertex;
      vertex[nVertex] = VertexInterp(isolevel,grid.p[1],grid.p[5],grid.val[1],grid.val[5],'Z');
      nVertex++;
   }
   if ((edgeTable[cubeindex] & 1024) == 1024){
      if (vertex.size() < nVertex+1 )
          vertex.resize( 2*vertex.size() );
      vertlist[10] = nVertex;
      vertex[nVertex] = VertexInterp(isolevel,grid.p[2],grid.p[6],grid.val[2],grid.val[6],'Z');
      nVertex++;
   }
   if ((edgeTable[cubeindex] & 2048) == 2048){
      if (vertex.size() < nVertex+1 )
          vertex.resize( 2*vertex.size() );
      vertlist[11] = nVertex;
      vertex[nVertex] = VertexInterp(isolevel,grid.p[3],grid.p[7],grid.val[3],grid.val[7],'Z');
      nVertex++;
   }
   cou[0] = nVertex;

   /* Create the Triangle */
   nTriang = cou[1];
   i = 0;
   for (i=0;triTable[cubeindex][i]!=-1;i+=3) {
       if (Triangles.size() < nTriang+1 )
           Triangles.resize( 2*Triangles.size() );
      Triangles[nTriang].p[0] = vertlist[triTable[cubeindex][i+2  ]];
      Triangles[nTriang].p[1] = vertlist[triTable[cubeindex][i+1]];
      Triangles[nTriang].p[2] = vertlist[triTable[cubeindex][i]];
      nTriang++;
   }
   cou[1] = nTriang;

   return 1;
}

/*
   Linearly interpolate the position where an isosurface cuts
   an edge between two vertices, each with their own scalar value
*/
static Xyz VertexInterp(double isolevel, Xyz p1, Xyz p2, double valp1, double valp2, char type)
{
   double mu;
   Xyz p;

   if (type == 'X'){
        p.y = p1.y; p.z = p1.z;
        p.x = (p2.x*valp1-p1.x*valp2)/(valp1-valp2);
   }
   else if (type == 'Y'){
       p.x = p1.x; p.z = p1.z;
       p.y = (p2.y*valp1-p1.y*valp2)/(valp1-valp2);
   }
   else if (type = 'Z'){
       p.x = p1.x; p.y = p1.y;
       p.z = (p2.z*valp1-p1.z*valp2)/(valp1-valp2);
   }
   return(p);
}

static void normalize(Xyz *normal){
    double norm = sqrt(normal->x*normal->x + normal->y*normal->y+normal->z+normal->z);
    if (norm > 0.00001){
        normal->x = normal->x/norm;
        normal->y = normal->y/norm;
        normal->z = normal->z/norm;
    }
}

int GLWidget::load_scene_file(const char *name) {
    FILE *fp = fopen(name, "r");
    if (!fp) return false;
    /*if (this->scene != NULL)
        close_gl(this->scene);*/

    lang_file(fp);
    if ((this->scene = scene_read()) == NULL)
        return false;
    fclose(fp);

    return true;
}

void init_lang(void)
{
  lang_defun((char *)("scene"), scene_parse);
  lang_defun((char *)("view"), view_parse);
  lang_defun((char *)("dist_light"), distlight_parse);
  lang_defun((char *)("plastic"), plastic_parse);

  lang_defun((char *)("shaderprog"), shaderprog_parse);

  lang_defun((char *)("meshobj"), obj_parse);
  lang_defun((char *)("mesh"), imesh_parse);

  lang_defun((char *)("primobj"), obj_parse);
  lang_defun((char *)("sphere"), sphere_parse);

  lang_defun((char *)("group"), graph_parse);
  lang_defun((char *)("affine"), xaffine_parse);

  lang_defun((char *)("slideproj"), slideproj_parse);
  lang_defun((char *)("shiny"), shiny_parse);
  lang_defun((char *)("rough"), rough_parse);
  lang_defun((char *)("textured"), textured_parse);
  lang_defun((char *)("imagemap"), imagemap_parse);
  lang_defun((char *)("chequer"), chequer_parse);

  lang_defun((char *)("arg"), arg_parse);
  lang_defun((char *)("motor"), motor_parse);
}

void scene_clone_gl(Object *ol1, Object *ol2)
{
  Object *o1, *o2;

  for (o1 = ol1, o2 = ol2; o1 != NULL || o2 != NULL; o1 = o1->next, o2 = o2->next) {
    if (o1->type == V_GROUP && o2->type == V_GROUP) {
      scene_clone_gl(children_(o1), children_(o2));
    } else {
      if (o1->u.sh->vb == NULL)
        o1->u.sh->vb = NEWSTRUCT(GLvbuf);
      *(o1->u.sh->vb) = *(o2->u.sh->vb);
      if (o1->mat)
        o1->mat->shd.id = o2->mat->shd.id;
      if (o2->mat->tinfo)
        *((TmapInfo *)(o1->mat->tinfo)) = *((TmapInfo *)(o2->mat->tinfo));
    }
  }
}
